﻿/*
Code written/modified by: Dean Lunz (aka Crated by: X)
Web: http://www.createdbyx.com/ 
NOTICE: The code in this file does not fall under the Zider licence.
  
 Original source for this code comes from the tutorial "Animation Sets" found on the
 http://www.codeanxiety.com/ website.
 Original Code Author: Justin "Kukyona" O'Dell -> http://codeanxiety.com/user/Profile.aspx?UserID=2100
 */
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace Zider.Animation
{
/// <summary>
/// AnimationSet will help you animate your sprites or models.
/// </summary>
public sealed class AnimationSet<FrameType>
{
    // set up variables and their default values
    private Dictionary<string, AnimationInfo> animations = new Dictionary<string, AnimationInfo>();
    private AnimationInfo currentAnimationInfo = null;
    private FrameType[] frames = null;
    public float frame = 0f;
    private int lastFrameIndex = 0;
    private int currentFrameIndex = 0;
    public bool animating = false;
    public bool animationFinished = false;

    /// <summary>
    /// Default constructor.
    /// </summary>
    /// <param name="frames">Frame data that we will be working with.</param>
    public AnimationSet( params FrameType[] frames )
    {
        this.frames = frames;

        // add a default animation that loops and encompasses all frames
        AddAnimation( "Default", new AnimationInfo( 0, frames.Length - 1, 15f, true ) );
        currentAnimationInfo = animations[ "Default" ];
    }

    /// <summary>
    /// Adds an animation to be used with this helper.
    /// </summary>
    /// <param name="name">The name of the animation.</param>
    /// <param name="startFrame">The index where this animation begins.</param>
    /// <param name="endFrame">The index where this animation ends.</param>
    /// <param name="loops">Sets weither this animation will start over when it reaches the end.</param>
    public void AddAnimation( string name, AnimationInfo info )
    {
        // if we already have an animation with this name then ignore the new one
        if( animations.ContainsKey( name ) )
            return;

        animations.Add( name, info );
    }

    /// <summary>
    /// Resets the current animation, if playing, and starts animating.
    /// </summary>
    public void Play()
    {
        // hitting play starts the animation from frame 0
        currentFrameIndex = currentAnimationInfo.StartFrame;
        frame = 0f;
        animationFinished = false;

        // only animate if we actually need to
        if( currentAnimationInfo.StartFrame != currentAnimationInfo.EndFrame )
            animating = true;
    }

    /// <summary>
    /// Starts playing the animation with the specified name.
    /// </summary>
    /// <param name="name">The name of the animation to play.</param>
    public void Play( string name )
    {
        // do we even have an animation with this name?
        if( animations.ContainsKey( name ) != true )
            return;

        // set the current animation
        currentAnimationInfo = animations[ name ];

        Play();
    }

    /// <summary>
    /// Pauses the animation at its current frame.
    /// </summary>
    public void Pause()
    {
        animating = false;
    }

    /// <summary>
    /// Resumes playing the current animation.
    /// </summary>
    public void Resume()
    {
        // only animate if we actually need to, single frame animations don't need to be animated at all
        if( currentAnimationInfo.StartFrame != currentAnimationInfo.EndFrame )
            animating = true;
    }

    /// <summary>
    /// Stop playing the current animation and resets it to the first frame.
    /// </summary>
    public void Stop()
    {
        // stopping resets the movie back to frame 0
        currentFrameIndex = currentAnimationInfo.StartFrame;
        frame = 0f;
        animating = false;
        animationFinished = false;
    }

    /// <summary>
    /// Called once per frame to help animate.
    /// </summary>
    /// <param name="gameTime">Snapshot of the current game timing.</param>
    public void Update( GameTime gameTime )
    {
        // if we aren't animating then theres nothing to do
        if( animating != true || currentAnimationInfo == null )
            return;

        // increment the current frame position
        frame += currentAnimationInfo.FrameRate * (float)gameTime.ElapsedGameTime.TotalSeconds;

        int totalFrames = currentAnimationInfo.EndFrame - currentAnimationInfo.StartFrame + 1;

        // have we hit our limit?
        if( (int)frame >= totalFrames )
        {
            // if the animation doesn't loop then just pause at the last frame
            if( currentAnimationInfo.Loops != true )
            {
                frame = (float)currentAnimationInfo.EndFrame;
                animating = false;
                animationFinished = true;
            }
            // otherwise, just move backwards to frame 0
            else
                // we subtract here that way if we were at frame 2.02 we'll start at frame 0.02
                frame -= (float)totalFrames;
        }

        // have we changed frames?
        if( (int)frame != currentFrameIndex )
        {
            lastFrameIndex = currentFrameIndex;
            currentFrameIndex = (int)frame;
        }
    }

    /// <summary>
    /// Gets the array of frames in this animation set.
    /// </summary>
    public FrameType[] Frames
    {
        get { return frames; }
    }

    /// <summary>
    /// Gets the current animation being played.
    /// </summary>
    public AnimationInfo CurrentAnimationInfo
    {
        get { return currentAnimationInfo; }
    }

    /// <summary>
    /// Gets whether or not the current animation has finished playing.
    /// </summary>
    public bool AnimationFinished
    {
        get { return animationFinished; }
    }

    /// <summary>
    /// Gets the current frame being played.
    /// </summary>
    public FrameType CurrentFrame
    {
        get { return frames[ currentFrameIndex ]; }
    }

    /// <summary>
    /// Gets the current frame index being played.
    /// </summary>
    public int CurrentFrameIndex
    {
        get { return currentFrameIndex; }
    }

    /// <summary>
    /// Gets the last frame that was played.
    /// </summary>
    public FrameType LastFrame
    {
        get { return frames[ lastFrameIndex ]; }
    }

    /// <summary>
    /// Gets the index of the last frame that was played.
    /// </summary>
    public int LastFrameIndex
    {
        get { return lastFrameIndex; }
    }

    /// <summary>
    /// Gets the ration of how far into the current frame we are in.
    /// 0.0 = 0%, 1.0 = 100%
    /// </summary>
    public float BlendRatio
    {
        get { return frame - (float)( (int)frame ); }
    }
}
}
