/*
Created by: X
http://www.createdbyx.com/

The Zider engine is licenced under the Creative Commons Attribution-NonCommercial-ShareAlike 2.5 Canada
http://creativecommons.org/licenses/by-nc-sa/2.5/ca/
*/
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Reflection;
using Zider.Properties;


namespace Zider
{
    public delegate bool GetTexture(string Name, out Texture2D tex);
    public static class Helpers
    {
        public const int EXITCODE_BadDefaultFont = -1;

        public static Rectangle CenterRectangle(Rectangle r, int x, int y)
        {
            return new Rectangle(x - (r.Width / 2), y - (r.Height / 2), r.Width, r.Height);
        }


        public static T FindComponent<T>(GameComponentCollection Items) where T : IGameComponent
        {
            for (int idx = 0; idx <= Items.Count - 1; idx++)
            {
                if (Items[idx].GetType() == typeof(T))
                {
                    return (T)Items[idx];
                }
            }

            return default(T);
        }

        public static List<IGameComponent> HoldComponents(ZiderGame game)
        {
            return HoldComponents(game, false);
        }


        public static List<DrawableGameComponent> SetVisibleComponents(ZiderGame game, bool value)
        {
            List<DrawableGameComponent> items = new List<DrawableGameComponent>();

            for (int idx = 0; idx <= game.Components.Count - 1; idx++)
            {
                object e;
                e = game.Components[idx];
                if (e is DrawableGameComponent)
                {
                    DrawableGameComponent dc = (DrawableGameComponent)e;
                    items.Add(dc);
                    dc.Visible = value;
                }
            }
            return items;
        }

        public static List<IGameComponent> HoldComponents(ZiderGame game, bool clear)
        {
            List<IGameComponent> items = new List<IGameComponent>();

            for (int idx = 0; idx <= game.Components.Count - 1; idx++)
            {
                items.Add(game.Components[idx]);
            }

            if (clear) { game.Components.Clear(); }

            return items;
        }

        public static void RestoreComponents(ZiderGame game, List<IGameComponent> items)
        {
            if (items == null) return;

            for (int idx = 0; idx <= items.Count - 1; idx++)
            {
                game.Components.Add(items[idx]);
            }
        }

        public static Classes.GUIWidget CreateWidget(ZiderGame ed, string file, string name)
        {
            return CreateWidget<Classes.GUIWidget>(ed, file, name);
        }

        public static T CreateWidget<T>(ZiderGame ed, string file, string name) where T : Classes.GUIWidget, new()
        {
            // get assembly ref
            Assembly assembly = Assembly.GetExecutingAssembly();
            // declare list to store controls
            List<T> items;
            // load controls
            items = xFripperyXNA.GUIReader.Load<T>(assembly.GetManifestResourceStream(file));
            // declare widget reference
            T w = null;
            // find widget, searching only top level widgets
            foreach (T e in items)
            {
                if (string.Compare(e.Name, name) == 0)
                {
                    w = e;
                    break;
                }
            }

            // check if found widget
            if (w == null)
            {
                throw new NullReferenceException(Resources.CouldNotFindWidget.Replace(Resources.RKEY, name) + "\n" +
                                                 Resources.SourceFile + file);
            }

            // assign the get texture callback
            SetGetTextureCallback(ed.GetGUITexture, w, true);

            // be sure to extract data info stord in the data property
            Helpers.ExtractData(w, true, file);

            // call init on widget
            w.Initilize(ed);

            // return wodget reference
            return w;
        }

        public static void SetGetTextureCallback(GetTexture gt, Classes.GUIWidget c, bool children)
        {
            c.GetTexture = gt;
            if (children)
            {
                foreach (Classes.GUIWidget gw in c.Children)
                {
                    SetGetTextureCallback(gt, gw, children);
                }
            }
        }

        public static void ExtractData(Classes.GUIWidget c, bool children)
        {
            ExtractData(c, children, string.Empty);
        }

        public static void ExtractData(Classes.GUIWidget c, bool children, string sourceFile)
        {
            System.Xml.XmlDocument doc = new System.Xml.XmlDocument();
            doc.LoadXml("<data>" + c.Data + "</data>");

            System.Xml.XmlNode node;
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_texture);

            // check if data contains a texture element
            if (node != null && node.InnerText != c.TextureName)
            {
                c.TextureName = node.InnerText;
            }


            Rectangle sr = c.SourceRectangle;
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_sourceleft);

            // extract source texture rect
            if (node != null && !int.TryParse(node.InnerText, out sr.X))
            {
                // do nothing?
            }
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_sourcetop);
            if (node != null && !int.TryParse(node.InnerText, out sr.Y))
            {
                // do nothing?
            }
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_sourcewidth);
            if (node != null && !int.TryParse(node.InnerText, out sr.Width))
            {
                // do nothing?
            }
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_sourceheight);
            if (node != null && !int.TryParse(node.InnerText, out sr.Height))
            {
                // do nothing?
            }
            // set source rect
            c.SourceRectangle = sr;

            // read color data
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_color);
            if (node != null)
            {
                Color col = c.Color;
                if (ParseColor(node, out col)) c.Color = col;
            }

            // check if data contains a text element
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_text);
            if (node != null) c.Text = node.InnerText;

            // process any scripting nodes
            ProcessScript(c.Script, sourceFile, doc.DocumentElement, true);

            // check if data contains a text alignment element
            node = doc.DocumentElement.SelectSingleNode(Resources.XML_textalignment);
            if (node != null)
            {

                Nuclex.Fonts.TextAlignment ta = Nuclex.Fonts.TextAlignment.Top | Nuclex.Fonts.TextAlignment.Left;
                try
                {
                    ta = (Nuclex.Fonts.TextAlignment)Enum.Parse(typeof(Nuclex.Fonts.TextAlignment), node.InnerText, true);
                }
                catch (Exception)
                {
                    throw new System.IO.InvalidDataException();
                }
                c.TextAlignment = ta;
            }

            doc = null;

            // check is wanting to extract child data as well
            if (children)
            {
                // extract data for children also
                foreach (Classes.GUIWidget child in c.Children)
                {
                    // if (child.GetType() == typeof(Interfaces.IGUIWidget))
                    // {
                    // Classes.GUIWidget gw = (GUIWidget)child;
                    // gw.ExtractData(children);
                    Helpers.ExtractData(child, children, sourceFile);
                    //}
                }
            }


        }

        public static void ProcessScript(System.Collections.Specialized.StringCollection script, 
                                         string sourceFile, System.Xml.XmlNode parent, 
                                         bool optimize)
        {
            // check if data contains a scriptfile element
            // NOTE: if a scriptfile element is specified and a script element is specified
            //       the sccriptfile element will be used as the script.
            System.Xml.XmlNode node = parent.SelectSingleNode(Resources.XML_scriptfile);
            string[] lines = null;
            if (node != null)
            {
                // get full file name from scriptfile element
                string file = node.InnerText;
                if (!System.IO.Path.IsPathRooted(file))
                {
                    string path = System.IO.Path.GetDirectoryName(sourceFile);
                    file = System.IO.Path.Combine(path, file);
                    file = System.IO.Path.GetFullPath(file);
                }
                // read all lines from the file
                lines = System.IO.File.ReadAllLines(file);
            }
            else
            {

                // check if data contains a script element
                node = parent.SelectSingleNode(Resources.XML_script);
                if (node != null)
                {
                    // get script lines from script node and remove empty lines
                    lines = node.InnerText.Split(new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries);
                }
            }

            // clear out any existing script
            script.Clear();
            // ad script lines
            if (lines != null) script.AddRange(lines);
            // remove comment lines and empty lines 
            if (optimize) OptimizeScript(script);
        }

        private static void OptimizeScript(System.Collections.Specialized.StringCollection script)
        {
            int idx = 0;
            // loop thru script lines
            while (idx <= script.Count - 1)
            {
                // check if line start with a script comment
                if (script[idx].StartsWith(Resources.ScriptComment))
                {
                    // remove line from script
                    script.RemoveAt(idx);
                }
                // check if line is empty
                else if (script[idx].Trim() == string.Empty)
                {
                    // remove line from script
                    script.RemoveAt(idx);
                }
                else
                {
                    // move onto next script line
                    script[idx] = script[idx].TrimStart();
                    idx++;
                }
            }
        }

        private static bool ParseColor(System.Xml.XmlNode node, out Color c)
        {
            c = Color.White;
            System.Xml.XmlNode rn, gn, bn, an;
            rn = node.Attributes.GetNamedItem(Resources.XML_r);
            gn = node.Attributes.GetNamedItem(Resources.XML_g);
            bn = node.Attributes.GetNamedItem(Resources.XML_b);
            an = node.Attributes.GetNamedItem(Resources.XML_a);

            byte rv = 255;
            byte gv = 255;
            byte bv = 255;
            byte av = 255; // set default alpha value

            // extract color components
            if (rn != null && !byte.TryParse(rn.InnerText, out rv))
            {
                return false;
            }
            if (gn != null && !byte.TryParse(gn.InnerText, out gv))
            {
                return false;
            }
            if (bn != null && !byte.TryParse(bn.InnerText, out bv))
            {
                return false;
            }
            if (an != null && !byte.TryParse(an.InnerText, out av))
            {
                // we can ignore alpha
            }

            // assign color
            c = new Color(rv, gv, bv, av);
            return true;
        }

        public static string ExtractData(string data, string e)
        {
            string tmp = "<" + e + ">";
            if (data.IndexOf(tmp) != -1)
            {
                data = data.Substring(data.IndexOf(tmp) + tmp.Length);
                data = data.Substring(0, data.IndexOf("<"));
                return data;
            }
            return string.Empty;
        }

        public static Dictionary<string, string> ExtractParameterData(string data, string e)
        {
            Dictionary<string, string> items;
            items = new Dictionary<string, string>();

            string tmp = "<" + e;
            if (data.IndexOf(tmp) != -1)
            {
                data = data.Substring(data.IndexOf(tmp) + tmp.Length);
                data = data.Substring(0, data.IndexOf(">")).Trim();
                if (data.EndsWith("/"))
                {
                    data = data.Remove(data.Length - 1).Trim();
                }

                while (data.Trim() != string.Empty)
                {
                    int pos;
                    string pName, pValue;
                    pos = data.IndexOf("=");
                    if (pos != -1)
                    {
                        pName = data.Substring(0, pos).Trim();
                        data = data.Substring(pos + 1);


                        pos = data.IndexOf("\"");
                        if (pos != -1)
                        {
                            int epos;
                            epos = data.IndexOf("\"", pos + 1);

                            if (epos != -1)
                            {
                                pValue = data.Substring(pos + 1, epos - (pos + 1)).Trim();
                                data = data.Substring(epos + 1);

                                // add entry
                                items.Add(pName, pValue);
                            }
                        }
                    }
                }
                return items;
            }
            return null;
        }

        public static Classes.GUIWidget ConvertToWidget(xFripperyXNA.Control c)
        {
            Classes.GUIWidget widget = new Classes.GUIWidget();
            widget.Anchor = c.Anchor;
            widget.Bounds = c.Bounds;
            widget.Children = c.Children;
            widget.Data = c.Data;
            widget.Name = c.Name;
            widget.Parent = c.Parent;
            widget.Visible = c.Visible;
            return widget;
        }

        public static Block CreateBlock(BlockType type, Rectangle source, string tex)
        {
            Block b;
            switch (type)
            {
                case BlockType.Platform:
                    b = new Classes.Platform();
                    break;

                default:
                    throw new System.IO.InvalidDataException(Resources.InvalidBlockType);

            }
            b.SourceRectangle = source;
            b.TextureName = tex;
            return b;
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="quat"></param>
        /// <param name="vector"></param>
        /// <returns></returns>
        public static Vector3 Vec3QuatMultiply(Quaternion quat, Vector3 vector)
        {
            // nVidia SDK implementation
            Vector3 uv, uuv;
            Vector3 qvec = new Vector3(quat.X, quat.Y, quat.Z);
            Vector3.Cross(ref qvec, ref vector, out uv);
            // uv = qvec.Cross(vector);
            Vector3.Cross(ref qvec, ref uv, out uuv);
            // uuv = qvec.Cross(uv);
            uv *= (2.0f * quat.W);
            uuv *= 2.0f;

            return vector + uv + uuv;

            // get the rotation matrix of the Quaternion and multiply it times the vector
            //return quat.ToRotationMatrix() * vector;
        }

    }
}
